#pragma once
#include "abc_datatype.h"

extern F32 fastlog(F32 x);
extern F32 sum_log_diag(rF32PTR p, rI32 K);
extern F32 sum_log_diagv2(rF32PTR p, rI32 K);
extern F32 sumlog(F32PTR p, I32 K);
extern F32 fastexp(F32 x);
extern F32 fastsqrt (F32 x);